#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

/* THESE MUST BE LAST of the sys defines */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "health.h"

/*Our globally visible stuffs, that we populate the global structs with */
static	int	health_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	health_probe(dev_info_t *dip);
static	int	health_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	health_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	health_print(dev_t, char *);

static	int	health_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	health_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	health_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	health_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp);
static	int	health_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

int debug_flag;
/* "char/block operations" */
/* maybe we need a bus_ops, too?!! */
static struct cb_ops	health_cb_ops = {
	health_open,		/* open() */
	health_close,		/* close() */
	nodev,			/* strategy()		*/
	health_print,		/* no print routine	*/
	nodev,			/* no dump routine	*/
	health_read,		/* read() */
	nodev,			/* write() */
	nodev,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	CB_REV,		/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops health_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	health_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	health_probe,		/* devo_probe */
	health_attach,		/* devo_attach */
	health_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&health_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in modctl.h
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"Solaris GerStra Health v0.1",
	&health_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 indicated by manual */
	(void *)&modldrv,
	NULL,			/* termination of list of linkage structures */
};

/**********************************************************************/
/* This is for use with soft_state_alloc(), etc.
 * This is a table ptr, which points to multiply-alloced things,
 * one for each instance of our driver.
 */
static void *health_globalstate;

static struct rawdata chipdata;

ddi_device_acc_attr_t dev_attributes={
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC /* dunno, so be strict */
};

/* This is the struct that define EVERY SINGLE TYPE OF "soft_state"
 * information stored about an instance of this driver!!
 */
struct health_state {
	int fakevar;
	ddi_acc_handle_t health_regs_map;
};

/**********************************************************************/

static void routineprint(char *string){
	if(debug_flag==0)
		return;
	cmn_err(CE_WARN, "health: In routine %s", string);
}
static void debugprint2(char *string){
	if(debug_flag==0)
		return;
	cmn_err(CE_NOTE, "%s", string);
}

static void printprop(dev_info_t *dip, char *propname){
	int val;
	val=ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0,propname, -1);
	cmn_err(CE_NOTE, "health: %s=%d", propname, val);
}


/* quickie from manpages */
/* we dont use it much :-) */
static kmutex_t   health_global_mutex;

/* why 5? dunno */

#define INITIAL_SOFT_SPACE 5

int
_init(void)
{
	int stateret;

	routineprint("_init");

	/*
	 * Initialize global mutex before mod_install'ing driver.
	 * If mod_install() fails, must clean up mutex initialization
	 */
	mutex_init(&health_global_mutex, "GerStra HEALTH Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	if ((stateret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&health_global_mutex);
		routineprint("init: modinstall !=0");
		debugprint2("health: _init FAILED");
		return stateret;
	}

	debugprint2("health: _init returning SUCCESS, supposedly");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	routineprint("_info");
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int       i;

	routineprint("_fini");
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((i = mod_remove(&modlinkage)) == 0) {
		mutex_destroy(&health_global_mutex);
	}
	return (i);
}

/************************************************************/
/* End of required "load driver" entry points */
/************************************************************/
/* baseaddr gets set to memory mapping of control register
 * regs_handle is just this funky thing the system needs
 */
caddr_t baseaddr;
ddi_acc_handle_t regs_handle;

/* return 1 if chip busy, 0 if not */
int chipbusy()
{
	uint16_t val;
	val=ddi_get16(regs_handle, (uint16_t*)baseaddr);
	if(val&0x808)
		return 1;
	return 0;
}

/* loops a bit if chip is busy. returns 0 when chip clear,
 * or -1 if chip always busy
 */
int waitforchip()
{
	if(chipbusy()){
		int count=1000;
		/*debugprint2("health: chip busy, retrying...");*/
		while(count>0){
			/* delay is .01 seconds */
			delay(drv_usectohz(100000));
			if(!chipbusy())
				break;
			count--;
		}
		if(count<0){
			return -1;
		}
	}
	return 0;
}

/* return status is 0 on okay, -1 on error */
/* -1 would be if chip is busy all the time */

int getchipdata(uint8_t reg, uint8_t *returnval)
{
	/* set "port" to right register location */
	if(waitforchip() ==-1)
	{
		return -1;
	}
	ddi_put8(regs_handle, baseaddr+8, reg);  /*select reg*/
	ddi_put8(regs_handle, baseaddr+4, 0x5A); /*and tell W83782D we want to write to 0x2D*/
	ddi_put8(regs_handle, baseaddr+2, 0x0A); /*by starting I2C cycle in byte mode*/
	if(waitforchip() ==-1)
	{
		return -1;
	}
	ddi_put8(regs_handle, baseaddr+4, 0x5B); /*tell W83782D we want to read from 0x2D*/
	ddi_put8(regs_handle, baseaddr+2, 0x0A); /*by starting I2C cycle in byte mode*/
	if(waitforchip() ==-1)
	{
		return -1;
	}
	*returnval=ddi_get8(regs_handle, baseaddr+6); /*finally get value*/
	return 0;
}

int getchipdataLM75(uint8_t reg, uint16_t *returnval)
{
	uint16_t a;
	if(waitforchip() ==-1)
	{
		return -1;
	}
	ddi_put8(regs_handle, baseaddr+8, reg);  /*select reg*/
	ddi_put8(regs_handle, baseaddr+4, 0x92); /*and tell W83782D we want to write to 0x49*/
	ddi_put8(regs_handle, baseaddr+2, 0x0A); /*by starting I2C cycle in byte mode*/
	if(waitforchip() ==-1)
	{
		return -1;
	}
	ddi_put8(regs_handle, baseaddr+4, 0x93); /*tell W83782D we want to read from 0x49*/
	ddi_put8(regs_handle, baseaddr+2, 0x0B); /*by starting I2C cycle in word mode*/
	if(waitforchip() ==-1)
	{
		return -1;
	}
	a=ddi_get16(regs_handle, (uint16_t*)(baseaddr+6)); /*finally get value*/
	*returnval=a&0x7FFF;
	return 0;
}

/* Sometimes, we just need to read the same "data" register again.
 * ugly damn architecture, but we're just doing what the chip designer
 * said...
 * (we always return 0)
*/
 int getchipnextdata(uint8_t *returnval)
{
	if(waitforchip() ==-1)
	{
		return -1;
	}
	ddi_put8(regs_handle, baseaddr+4, 0x5B); /*tell W83782D we want to read from 0x2D*/
	ddi_put8(regs_handle, baseaddr+2, 0x0A); /*by starting I2C cycle in byte mode*/
	if(waitforchip() ==-1)
	{
		return -1;
	}
	*returnval=ddi_get8(regs_handle, baseaddr+6); /*finally get value*/
	return 0;
}



/* set special chip-internal register stuff.
 * up to CALLER to avoid reentrancy problems
 * return 0 on okay, -1 on fail
 */
int setchipdata(uint8_t reg, uint8_t newvalue)
{}
/*{
 *	if(waitforchip() ==-1){
 *		return -1;
 *	}
 *	
 *	ddi_put8(regs_handle, baseaddr, reg);
 *	ddi_put8(regs_handle, baseaddr+1, newvalue);
 *	return 0;
 *}
 */

/*********************************************************************/
/* here are the standard entry points we configured in health_cb_ops */
/*********************************************************************/

static int
health_print(dev_t dev, char *string)
{
	int instance = getminor(dev);
	/*health_devstate_t *rsp = ddi_get_soft_state(health_state, instance);*/
	cmn_err(CE_WARN, "%s(%d): %s\n", "health", instance, string);

	return (0);
}

static	int	health_strategy(struct buf* b){
	routineprint("strategy");

	/* remember to call biodone() before exit!!! */

	return (ENXIO);
   	
}

static int	health_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	routineprint("getinfo, kinda faked");
	if(infocmd== DDI_INFO_DEVT2INSTANCE){
		*result=(void*) getminor((dev_t) arg);
		return DDI_SUCCESS;
	}
	/* we're actually supposed to ignore the passed in
	 * dip pointer, according to the manpage
	 */
	*result=dip;
	return DDI_SUCCESS;
}


/* We assume the system will ensure non-reentrancy here */
static	int health_probe(dev_info_t *dip){
    	short id;
	uint8_t byteval;
	int intid;
	int retval;
	caddr_t tmpcntl;
	ddi_acc_handle_t tmp_handle;

   	routineprint("probe");

	/* both tmpcntl, AND tmp_handle get set. huh?!! */
	retval=ddi_regs_map_setup(dip, 0, &tmpcntl,
			      0,0,
			      & dev_attributes, &tmp_handle
			      );
	if(retval !=  DDI_SUCCESS){
		if(retval==DDI_FAILURE){
			debugprint2("health: mapping returned FAILURE");
		} else {
			debugprint2("health: mapping failed.(nonfail)");
		}
		return DDI_PROBE_FAILURE;
	}


/*
 * This is where we're supposed to check if our hardware really is present
 * on the system.
 */


  	debugprint2("health: full probe succeeded!");
	/* let's don't and say we did :-) */
	return DDI_PROBE_SUCCESS;
}

static	int	health_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
	ddi_device_acc_attr_t dev_attributes;
	int instance;
	int val; /* generic temp value */

	instance = ddi_get_instance(dip);

	debug_flag=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
				    "debug-level", 0);

	routineprint("attach");

   	switch(cmd){
	 case DDI_RESUME:
		debugprint2("health: Ignoring attach_RESUME");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		debugprint2("health: Ignoring attach_PM_RESUME");
		return DDI_FAILURE;
	 case DDI_ATTACH:
		/* hardcode this bit, sigh*/
		dev_attributes.devacc_attr_version = 1;
		dev_attributes.devacc_attr_dataorder = DDI_STRICTORDER_ACC;
		dev_attributes.devacc_attr_endian_flags = DDI_STRUCTURE_LE_ACC;
		
		
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "health: unknown attach command\n");
	   	return DDI_FAILURE;
	}

	val=ddi_regs_map_setup(dip, 0, &baseaddr,
			      0,0,
			      & dev_attributes, &regs_handle
			      );
	if(val !=  DDI_SUCCESS){
		if(val==DDI_FAILURE){
			debugprint2("health: mapping returned FAILURE");
		} else {
			debugprint2("health: mapping failed.(nonfail)");
		}
		return DDI_FAILURE;
	}

	debugprint2("health: Attach GOOD!");

	bcopy(chipdata.chipid, "W83782D\0",8);

   	ddi_create_minor_node(dip, "", S_IFCHR, instance,
			      DDI_PSEUDO, 0);

	return DDI_SUCCESS;
	/*return DDI_FAILURE;*/
}

static	int	health_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	uint8_t throwaway, accum1;
	uint16_t accum16;
   	routineprint("open");

	mutex_enter(&health_global_mutex);

	/*
	* we need to do a read of chip fvalues as they are NOW,
	* and update chipdata
	*/
	if(waitforchip()==-1)
	{
		debugprint2("health: chip too busy");
		mutex_exit(&health_global_mutex);
		return (ENXIO);
	}
	getchipdataLM75(0,&accum16);
	chipdata.cputemp[1]=accum16;

	getchipdata(0x29, &chipdata.fanrpm);
	
/*	getchipdata(0x47, &accum1);
 *	chipdata.fandiv=(accum1>>6) & 0x03;
 */
	chipdata.fandiv=1;

	getchipdata(0x22, &chipdata.volts[2]);
	getchipdata(0x23, &chipdata.volts[3]);
	getchipdata(0x24, &chipdata.volts[4]);
	getchipdata(0x27, &accum1);
	chipdata.cputemp[0]= (int)(accum1);

	mutex_exit(&health_global_mutex);
   	/*return (ENXIO);*/
   	return 0;
}



/* normally devices should check for perms, minor num, etc.
 * but we dont care about that right now
 * we're bad,and only look at the first iovec.
 * The chipdata structure is updated by the
 * open() call. 
 */
static	int	health_read(dev_t dev, struct uio* ustr, cred_t * credp){
	void *dataptr=&chipdata;
	int datalen=sizeof(chipdata);
	int copylen;
	int dataoffset;
	iovec_t *iovec_ptr;

   	routineprint("read");

	iovec_ptr=ustr->uio_iov;
	if(ustr->uio_segflg != UIO_USERSPACE){
		debugprint2("health_read from non-user space?");
		return;
	}
	if(ustr->uio_iovcnt >1){
		debugprint2("health_read given multiple iovecs?");
	}

	dataoffset=ustr->uio_offset;
	if(dataoffset >=datalen){
		if(debug_flag)
			cmn_err(CE_WARN, "health_read overflow: offset=%d",
				dataoffset);
		/* This is not documented, but apparently, it is
		 * normal to just return "0 bytes read" in these
		 * situations
		 */
		dataoffset=0;
		copylen=0;
	} else {

		/* silently truncate reads that start IN data, but
		 * request past end of our datablock
		 */
		copylen = datalen - dataoffset;
	}

	/* "normal" thing for most kernel read() routines is to call
	 * physio(udma_strategy, NULL, dev, B_READ,
	 *	  minphys[system routine], ustr);
	 * but this works for our purposes.
	 */
	if(copylen > iovec_ptr->iov_len){
		copylen = iovec_ptr->iov_len;
	}

	if(debug_flag){
		cmn_err(CE_NOTE, "health: start %d, size %d",
			dataoffset, copylen);
	}

	if(uiomove(dataptr+dataoffset, copylen, UIO_READ, ustr) != 0){
		return DDI_FAILURE;
	}
	return DDI_SUCCESS;
}

static	int	health_write(dev_t dev, struct uio *uiop, cred_t *credp){
   	routineprint("write");
	/* /dev/null, so to speak */
   	return (0);
}

static	int	health_close(dev_t dev, int openflags, int otyp, cred_t *credp){
   	routineprint("close (faked)");
   	return (0);
}

static	int	health_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){
   	routineprint("ioctl");
   	return (ENXIO);
}


/* clean up anything allocated in health_attach */
static	int	health_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
	int instance=ddi_get_instance(dip);
   	routineprint("detach");

	if(cmd != DDI_DETACH){
		debugprint2("Ignoring non-detach call to detach\n");
		return DDI_SUCCESS;
	}

	ddi_regs_map_free(&regs_handle);

	return (DDI_SUCCESS);
		     
}

